import os.path
import tempfile

import boto3
import botocore
import cdblib
import pytest
from moto import mock_s3

from leveled_hotbackup_s3_sync import erlang
from leveled_hotbackup_s3_sync.utils import (
    RiakObject,
    check_endpoint_url,
    check_s3_url,
    create_journal_key,
    download_bytes_from_s3,
    download_file_from_s3,
    find_latest_ring,
    find_primary_partition,
    get_owned_partitions,
    get_ring_size,
    hash_bucket_key,
    is_s3_url,
    local_path_exists,
    parse_s3_url,
    riak_ring_increment,
    riak_ring_indexes,
    s3_path_exists,
    str_to_bytes,
    swap_path,
    upload_bytes_to_s3,
    upload_file_to_s3,
)

TEST_RING_DIRECTORY = "/tmp/a5017381-4c3e-46e6-bd02-342c4b894b59-ring"
TEST_RING_FILENAME = "/tmp/a5017381-4c3e-46e6-bd02-342c4b894b59-ring/riak_core_ring.default.20240116160656"
TEST_RING_FILENAME2 = "/tmp/a5017381-4c3e-46e6-bd02-342c4b894b59-ring/riak_core_ring.default.20231025141013"


@pytest.fixture(name="s3_client")
def fixture_s3_client():
    with mock_s3():
        s3_client = boto3.client("s3", region_name="us-east-1")
        s3_client.create_bucket(Bucket="test")
        s3_client.put_bucket_versioning(
            Bucket="test",
            VersioningConfiguration={
                "MFADelete": "Disabled",
                "Status": "Enabled",
            },
        )
        yield s3_client


def test_str_to_bytes():
    assert str_to_bytes("test") == b"test"
    assert str_to_bytes("\u0100") == b"\xc4\x80"


def test_is_s3_url():
    assert is_s3_url("s3://test/path/key") is True
    assert is_s3_url("/file/path") is False


def test_check_s3_url():
    assert check_s3_url("s3://test/path/key") == "s3://test/path/key"
    with pytest.raises(ValueError):
        check_s3_url("randomtext")
    with pytest.raises(ValueError):
        check_s3_url("/file/path")
    with pytest.raises(ValueError):
        check_s3_url("https://www.google.com")


def test_check_endpoint_url():
    assert check_endpoint_url("http://localhost") == "http://localhost"
    with pytest.raises(ValueError):
        check_endpoint_url("http://localhost/path")


def test_s3_path_exists(s3_client):
    s3_client.put_object(Bucket="test", Key="testkey", Body=b"test")
    s3_client.put_object(Bucket="test", Key="path/to/testkey", Body=b"test")

    assert s3_path_exists("s3://test/testkey", None) is True
    assert s3_path_exists("s3://test/path/to/testkey", None) is True
    assert s3_path_exists("s3://test/doesntexist", None) is False
    with pytest.raises(botocore.exceptions.ClientError) as exc:
        s3_path_exists("s3://nobucket/key", None)
    err = exc.value.response["Error"]
    assert err["Message"] == "The specified bucket does not exist"


def test_upload_file_to_s3(s3_client):
    with tempfile.NamedTemporaryFile() as file_handle:
        file_handle.write(b"Hello world!")
        file_handle.flush()
        upload_file_to_s3(file_handle.name, "s3://test/hello", None)

    s3_obj = s3_client.get_object(Bucket="test", Key="hello")
    assert s3_obj["Body"].read() == b"Hello world!"


def test_upload_bytes_to_s3(s3_client):
    upload_bytes_to_s3(b"Hello world!", "s3://test/world", None)
    s3_obj = s3_client.get_object(Bucket="test", Key="world")
    assert s3_obj["Body"].read() == b"Hello world!"


def test_download_file_from_s3(s3_client):
    s3_client.put_object(Bucket="test", Key="testdownloadfile", Body=b"testdownload")
    with tempfile.NamedTemporaryFile() as file_handle:
        download_file_from_s3("s3://test/testdownloadfile", file_handle.name, None)
        with open(file_handle.name, "rb") as downloaded_handle:
            assert downloaded_handle.read() == b"testdownload"


def test_download_bytes_from_s3(s3_client):
    s3_client.put_object(Bucket="test", Key="testdownloadbytes", Body=b"testbytes")
    assert download_bytes_from_s3("s3://test/testdownloadbytes", None) == b"testbytes"

    v1_response = s3_client.put_object(Bucket="test", Key="testdownloadbytesversion", Body=b"testbytesv1")
    v2_response = s3_client.put_object(Bucket="test", Key="testdownloadbytesversion", Body=b"testbytesv2")

    assert v1_response["VersionId"] != v2_response["VersionId"]

    assert (
        download_bytes_from_s3("s3://test/testdownloadbytesversion", None, v1_response["VersionId"]) == b"testbytesv1"
    )
    assert (
        download_bytes_from_s3("s3://test/testdownloadbytesversion", None, v2_response["VersionId"]) == b"testbytesv2"
    )


def test_local_path_exists():
    assert local_path_exists(os.path.curdir) is True
    assert local_path_exists("/this/path/shouldnt/exist") is False


def test_parse_s3_url():
    bucket, bkey = parse_s3_url("s3://testbucket/testpath/testkey")
    assert bucket == "testbucket"
    assert bkey == "testpath/testkey"

    with pytest.raises(ValueError):
        parse_s3_url("https://testbucket/testpath/testkey")

    with pytest.raises(ValueError):
        parse_s3_url("randomtextiswrong")


def test_swap_path():
    assert swap_path("/source/path/file.name", "/source/path", "/new/dest/path") == "/new/dest/path/file.name"


def test_hash_bucket_key():
    assert hash_bucket_key(b"testBucket", b"testKey") == 1419470377952330688790609328442920717178450161891
    assert hash_bucket_key(b"testBucket", b"testKey9922") == 1449677501307210311800392832900504468941568173924
    assert hash_bucket_key(b"anotherbucket", b"k") == 1162606229195745701566405011540602585179415508602
    assert hash_bucket_key(b"bucketType", b"k", b"typedBucket") == 1294252316610983544584872785443701753478726932485


def test_riak_ring_increment():
    assert riak_ring_increment(32) == 45671926166590716193865151022383844364247891967
    assert riak_ring_increment(128) == 11417981541647679048466287755595961091061972991
    assert riak_ring_increment(512) == 2854495385411919762116571938898990272765493247


def test_riak_ring_indexes():
    assert riak_ring_indexes(32) == RING_SIZE_32
    assert riak_ring_indexes(128) == RING_SIZE_128
    assert riak_ring_indexes(512) == RING_SIZE_512


def test_find_primary_partition():
    assert find_primary_partition(32, b"testBucket", b"testKey") == 0
    assert find_primary_partition(32, b"testBucket", b"testKey9922") == 0
    assert find_primary_partition(32, b"anotherbucket", b"k") == 1187470080331358621040493926581979953470445191168
    assert (
        find_primary_partition(32, b"typedBucket", b"k", b"bucketType")
        == 1415829711164312202009819681693899175291684651008
    )
    assert find_primary_partition(128, b"testBucket", b"testKey") == 1427247692705959881058285969449495136382746624000
    assert (
        find_primary_partition(128, b"testBucket", b"testKey9922") == 1450083655789255239155218544960687058564870569984
    )
    assert find_primary_partition(128, b"anotherbucket", b"k") == 1164634117248063262943561351070788031288321245184
    assert (
        find_primary_partition(128, b"typedBucket", b"k", b"bucketType")
        == 1381575766539369164864420818427111292018498732032
    )


def test_create_journal_key():
    filename = (
        "/tmp/a5017381-4c3e-46e6-bd02-342c4b894b59/0/journal/journal_files/0_50f4666b-6ad8-4b6f-9e2a-23a235c82706.cdb"
    )
    with cdblib.Reader.from_file_path(filename) as reader:
        for journal_key_bin in reader.keys():
            journal_key = erlang.binary_to_term(journal_key_bin)
            buckettype = None
            if isinstance(journal_key[2][1], tuple):
                buckettype = journal_key[2][1][0].value
                bucket = journal_key[2][1][1].value
            else:
                bucket = journal_key[2][1].value
            assert journal_key_bin == create_journal_key(journal_key[0], bucket, journal_key[2][2].value, buckettype)

    filename = (
        "/tmp/a5017381-4c3e-46e6-bd02-342c4b894b59/0/journal/journal_files/972_e6205c6c-3b8b-40e6-baee-295dcc76488a.cdb"
    )
    with cdblib.Reader.from_file_path(filename) as reader:
        for journal_key_bin in reader.keys():
            journal_key = erlang.binary_to_term(journal_key_bin)
            buckettype = None
            if isinstance(journal_key[2][1], tuple):
                buckettype = journal_key[2][1][0].value
                bucket = journal_key[2][1][1].value
            else:
                bucket = journal_key[2][1].value
            assert journal_key_bin == create_journal_key(journal_key[0], bucket, journal_key[2][2].value, buckettype)


# pylint: disable=protected-access
class TestRiakObject:
    def test_decode_maybe_binary(self):
        riak_object = RiakObject()
        assert riak_object._decode_maybe_binary(b"\x01test") == b"test"
        test_val = erlang.term_to_binary(["this", "is", "a", "test"])
        assert riak_object._decode_maybe_binary(b"\x00" + test_val) == [b"this", b"is", b"a", b"test"]

    def test_decode(self):
        riak_object = RiakObject()
        riak_object.decode(
            b"5\x01\x00\x00\x00&\x83l\x00\x00\x00\x01h\x02m\x00\x00\x00\x0c\xbf\x00\xa1\xef\x9e\xc4\xd4\xfe\x00\x00"
            b'\x00\x14h\x02a\x02n\x05\x00\xff\xb0\x97\xdd\x0ej\x00\x00\x00\x01\x00\x00\x00\x17\x01{"test":"replaced'
            b'992"}\x00\x00\x00\x87\x00\x00\x06\x90\x00\x00p\xff\x00\x07\xd5\xfd\x166qFSOBD6ZxTXWCznwiCdM8\x00\x00'
            b"\x00\x00\x0c\x01X-Riak-Meta\x00\x00\x00\x03\x00\x83j\x00\x00\x00\x06\x01index\x00\x00\x00\x03\x00\x83"
            b"j\x00\x00\x00\r\x01content-type\x00\x00\x00\x15\x00\x83k\x00\x10application/json\x00\x00\x00\x06\x01"
            b"Links\x00\x00\x00\x03\x00\x83j"
        )
        assert riak_object.vector_clocks == [
            (
                erlang.OtpErlangBinary(b"\xbf\x00\xa1\xef\x9e\xc4\xd4\xfe\x00\x00\x00\x14", bits=8),
                (2, 63847248127),
            )
        ]
        assert riak_object.siblings == [
            {
                "value": b'{"test":"replaced992"}',
                "metadata": {
                    "last_modified": "1680028927.513533",
                    "vtag": b"6qFSOBD6ZxTXWCznwiCdM8",
                    "deleted": 0,
                    "X-Riak-Meta": [],
                    "index": [],
                    "content-type": b"application/json",
                    "Links": [],
                },
            }
        ]

    def test_decode_wrong_magic_number(self):
        riak_object = RiakObject()
        with pytest.raises(ValueError) as exc:
            riak_object.decode(b"wontwork")
        assert str(exc.value) == "Decode error, wrong magic number"

    def test_decode_wrong_object_version(self):
        riak_object = RiakObject()
        with pytest.raises(ValueError) as exc:
            riak_object.decode(b"5\x02\x00\x00\x00")
        assert str(exc.value) == "Decode error, wrong object version"

    def test_decode_unexpected_data(self):
        riak_object = RiakObject()
        with pytest.raises(ValueError) as exc:
            riak_object.decode(
                b"5\x01\x00\x00\x00&\x83l\x00\x00\x00\x01h\x02m\x00\x00\x00\x0c\xbf\x00\xa1\xef\x9e\xc4\xd4\xfe\x00\x00"
                b'\x00\x14h\x02a\x02n\x05\x00\xff\xb0\x97\xdd\x0ej\x00\x00\x00\x01\x00\x00\x00\x17\x01{"test":"replaced'
                b'992"}\x00\x00\x00\x87\x00\x00\x06\x90\x00\x00p\xff\x00\x07\xd5\xfd\x166qFSOBD6ZxTXWCznwiCdM8\x00\x00'
                b"\x00\x00\x0c\x01X-Riak-Meta\x00\x00\x00\x03\x00\x83j\x00\x00\x00\x06\x01index\x00\x00\x00\x03\x00\x83"
                b"j\x00\x00\x00\r\x01content-type\x00\x00\x00\x15\x00\x83k\x00\x10application/json\x00\x00\x00\x06\x01"
                b"Links\x00\x00\x00\x03\x00\x83j\x00"
            )
        assert str(exc.value) == "Decode error, unexpected data"


def test_find_latest_ring():
    with pytest.raises(ValueError) as exc:
        find_latest_ring("/tmp")
    assert str(exc.value) == "/tmp is not a valid Riak Ring location"
    assert find_latest_ring(TEST_RING_DIRECTORY) == TEST_RING_FILENAME


def test_get_ring_size():
    assert get_ring_size(TEST_RING_FILENAME) == 64


def test_get_owned_partitions():
    assert get_owned_partitions(TEST_RING_FILENAME) == riak_ring_indexes(64)
    partitions = get_owned_partitions(TEST_RING_FILENAME2)
    assert len(partitions) == 51


RING_SIZE_32 = [
    0,
    45671926166590716193865151022383844364247891968,
    91343852333181432387730302044767688728495783936,
    137015778499772148581595453067151533092743675904,
    182687704666362864775460604089535377456991567872,
    228359630832953580969325755111919221821239459840,
    274031556999544297163190906134303066185487351808,
    319703483166135013357056057156686910549735243776,
    365375409332725729550921208179070754913983135744,
    411047335499316445744786359201454599278231027712,
    456719261665907161938651510223838443642478919680,
    502391187832497878132516661246222288006726811648,
    548063113999088594326381812268606132370974703616,
    593735040165679310520246963290989976735222595584,
    639406966332270026714112114313373821099470487552,
    685078892498860742907977265335757665463718379520,
    730750818665451459101842416358141509827966271488,
    776422744832042175295707567380525354192214163456,
    822094670998632891489572718402909198556462055424,
    867766597165223607683437869425293042920709947392,
    913438523331814323877303020447676887284957839360,
    959110449498405040071168171470060731649205731328,
    1004782375664995756265033322492444576013453623296,
    1050454301831586472458898473514828420377701515264,
    1096126227998177188652763624537212264741949407232,
    1141798154164767904846628775559596109106197299200,
    1187470080331358621040493926581979953470445191168,
    1233142006497949337234359077604363797834693083136,
    1278813932664540053428224228626747642198940975104,
    1324485858831130769622089379649131486563188867072,
    1370157784997721485815954530671515330927436759040,
    1415829711164312202009819681693899175291684651008,
]

RING_SIZE_128 = [
    0,
    11417981541647679048466287755595961091061972992,
    22835963083295358096932575511191922182123945984,
    34253944624943037145398863266787883273185918976,
    45671926166590716193865151022383844364247891968,
    57089907708238395242331438777979805455309864960,
    68507889249886074290797726533575766546371837952,
    79925870791533753339264014289171727637433810944,
    91343852333181432387730302044767688728495783936,
    102761833874829111436196589800363649819557756928,
    114179815416476790484662877555959610910619729920,
    125597796958124469533129165311555572001681702912,
    137015778499772148581595453067151533092743675904,
    148433760041419827630061740822747494183805648896,
    159851741583067506678528028578343455274867621888,
    171269723124715185726994316333939416365929594880,
    182687704666362864775460604089535377456991567872,
    194105686208010543823926891845131338548053540864,
    205523667749658222872393179600727299639115513856,
    216941649291305901920859467356323260730177486848,
    228359630832953580969325755111919221821239459840,
    239777612374601260017792042867515182912301432832,
    251195593916248939066258330623111144003363405824,
    262613575457896618114724618378707105094425378816,
    274031556999544297163190906134303066185487351808,
    285449538541191976211657193889899027276549324800,
    296867520082839655260123481645494988367611297792,
    308285501624487334308589769401090949458673270784,
    319703483166135013357056057156686910549735243776,
    331121464707782692405522344912282871640797216768,
    342539446249430371453988632667878832731859189760,
    353957427791078050502454920423474793822921162752,
    365375409332725729550921208179070754913983135744,
    376793390874373408599387495934666716005045108736,
    388211372416021087647853783690262677096107081728,
    399629353957668766696320071445858638187169054720,
    411047335499316445744786359201454599278231027712,
    422465317040964124793252646957050560369293000704,
    433883298582611803841718934712646521460354973696,
    445301280124259482890185222468242482551416946688,
    456719261665907161938651510223838443642478919680,
    468137243207554840987117797979434404733540892672,
    479555224749202520035584085735030365824602865664,
    490973206290850199084050373490626326915664838656,
    502391187832497878132516661246222288006726811648,
    513809169374145557180982949001818249097788784640,
    525227150915793236229449236757414210188850757632,
    536645132457440915277915524513010171279912730624,
    548063113999088594326381812268606132370974703616,
    559481095540736273374848100024202093462036676608,
    570899077082383952423314387779798054553098649600,
    582317058624031631471780675535394015644160622592,
    593735040165679310520246963290989976735222595584,
    605153021707326989568713251046585937826284568576,
    616571003248974668617179538802181898917346541568,
    627988984790622347665645826557777860008408514560,
    639406966332270026714112114313373821099470487552,
    650824947873917705762578402068969782190532460544,
    662242929415565384811044689824565743281594433536,
    673660910957213063859510977580161704372656406528,
    685078892498860742907977265335757665463718379520,
    696496874040508421956443553091353626554780352512,
    707914855582156101004909840846949587645842325504,
    719332837123803780053376128602545548736904298496,
    730750818665451459101842416358141509827966271488,
    742168800207099138150308704113737470919028244480,
    753586781748746817198774991869333432010090217472,
    765004763290394496247241279624929393101152190464,
    776422744832042175295707567380525354192214163456,
    787840726373689854344173855136121315283276136448,
    799258707915337533392640142891717276374338109440,
    810676689456985212441106430647313237465400082432,
    822094670998632891489572718402909198556462055424,
    833512652540280570538039006158505159647524028416,
    844930634081928249586505293914101120738586001408,
    856348615623575928634971581669697081829647974400,
    867766597165223607683437869425293042920709947392,
    879184578706871286731904157180889004011771920384,
    890602560248518965780370444936484965102833893376,
    902020541790166644828836732692080926193895866368,
    913438523331814323877303020447676887284957839360,
    924856504873462002925769308203272848376019812352,
    936274486415109681974235595958868809467081785344,
    947692467956757361022701883714464770558143758336,
    959110449498405040071168171470060731649205731328,
    970528431040052719119634459225656692740267704320,
    981946412581700398168100746981252653831329677312,
    993364394123348077216567034736848614922391650304,
    1004782375664995756265033322492444576013453623296,
    1016200357206643435313499610248040537104515596288,
    1027618338748291114361965898003636498195577569280,
    1039036320289938793410432185759232459286639542272,
    1050454301831586472458898473514828420377701515264,
    1061872283373234151507364761270424381468763488256,
    1073290264914881830555831049026020342559825461248,
    1084708246456529509604297336781616303650887434240,
    1096126227998177188652763624537212264741949407232,
    1107544209539824867701229912292808225833011380224,
    1118962191081472546749696200048404186924073353216,
    1130380172623120225798162487804000148015135326208,
    1141798154164767904846628775559596109106197299200,
    1153216135706415583895095063315192070197259272192,
    1164634117248063262943561351070788031288321245184,
    1176052098789710941992027638826383992379383218176,
    1187470080331358621040493926581979953470445191168,
    1198888061873006300088960214337575914561507164160,
    1210306043414653979137426502093171875652569137152,
    1221724024956301658185892789848767836743631110144,
    1233142006497949337234359077604363797834693083136,
    1244559988039597016282825365359959758925755056128,
    1255977969581244695331291653115555720016817029120,
    1267395951122892374379757940871151681107879002112,
    1278813932664540053428224228626747642198940975104,
    1290231914206187732476690516382343603290002948096,
    1301649895747835411525156804137939564381064921088,
    1313067877289483090573623091893535525472126894080,
    1324485858831130769622089379649131486563188867072,
    1335903840372778448670555667404727447654250840064,
    1347321821914426127719021955160323408745312813056,
    1358739803456073806767488242915919369836374786048,
    1370157784997721485815954530671515330927436759040,
    1381575766539369164864420818427111292018498732032,
    1392993748081016843912887106182707253109560705024,
    1404411729622664522961353393938303214200622678016,
    1415829711164312202009819681693899175291684651008,
    1427247692705959881058285969449495136382746624000,
    1438665674247607560106752257205091097473808596992,
    1450083655789255239155218544960687058564870569984,
]

RING_SIZE_512 = [
    0,
    2854495385411919762116571938898990272765493248,
    5708990770823839524233143877797980545530986496,
    8563486156235759286349715816696970818296479744,
    11417981541647679048466287755595961091061972992,
    14272476927059598810582859694494951363827466240,
    17126972312471518572699431633393941636592959488,
    19981467697883438334816003572292931909358452736,
    22835963083295358096932575511191922182123945984,
    25690458468707277859049147450090912454889439232,
    28544953854119197621165719388989902727654932480,
    31399449239531117383282291327888893000420425728,
    34253944624943037145398863266787883273185918976,
    37108440010354956907515435205686873545951412224,
    39962935395766876669632007144585863818716905472,
    42817430781178796431748579083484854091482398720,
    45671926166590716193865151022383844364247891968,
    48526421552002635955981722961282834637013385216,
    51380916937414555718098294900181824909778878464,
    54235412322826475480214866839080815182544371712,
    57089907708238395242331438777979805455309864960,
    59944403093650315004448010716878795728075358208,
    62798898479062234766564582655777786000840851456,
    65653393864474154528681154594676776273606344704,
    68507889249886074290797726533575766546371837952,
    71362384635297994052914298472474756819137331200,
    74216880020709913815030870411373747091902824448,
    77071375406121833577147442350272737364668317696,
    79925870791533753339264014289171727637433810944,
    82780366176945673101380586228070717910199304192,
    85634861562357592863497158166969708182964797440,
    88489356947769512625613730105868698455730290688,
    91343852333181432387730302044767688728495783936,
    94198347718593352149846873983666679001261277184,
    97052843104005271911963445922565669274026770432,
    99907338489417191674080017861464659546792263680,
    102761833874829111436196589800363649819557756928,
    105616329260241031198313161739262640092323250176,
    108470824645652950960429733678161630365088743424,
    111325320031064870722546305617060620637854236672,
    114179815416476790484662877555959610910619729920,
    117034310801888710246779449494858601183385223168,
    119888806187300630008896021433757591456150716416,
    122743301572712549771012593372656581728916209664,
    125597796958124469533129165311555572001681702912,
    128452292343536389295245737250454562274447196160,
    131306787728948309057362309189353552547212689408,
    134161283114360228819478881128252542819978182656,
    137015778499772148581595453067151533092743675904,
    139870273885184068343712025006050523365509169152,
    142724769270595988105828596944949513638274662400,
    145579264656007907867945168883848503911040155648,
    148433760041419827630061740822747494183805648896,
    151288255426831747392178312761646484456571142144,
    154142750812243667154294884700545474729336635392,
    156997246197655586916411456639444465002102128640,
    159851741583067506678528028578343455274867621888,
    162706236968479426440644600517242445547633115136,
    165560732353891346202761172456141435820398608384,
    168415227739303265964877744395040426093164101632,
    171269723124715185726994316333939416365929594880,
    174124218510127105489110888272838406638695088128,
    176978713895539025251227460211737396911460581376,
    179833209280950945013344032150636387184226074624,
    182687704666362864775460604089535377456991567872,
    185542200051774784537577176028434367729757061120,
    188396695437186704299693747967333358002522554368,
    191251190822598624061810319906232348275288047616,
    194105686208010543823926891845131338548053540864,
    196960181593422463586043463784030328820819034112,
    199814676978834383348160035722929319093584527360,
    202669172364246303110276607661828309366350020608,
    205523667749658222872393179600727299639115513856,
    208378163135070142634509751539626289911881007104,
    211232658520482062396626323478525280184646500352,
    214087153905893982158742895417424270457411993600,
    216941649291305901920859467356323260730177486848,
    219796144676717821682976039295222251002942980096,
    222650640062129741445092611234121241275708473344,
    225505135447541661207209183173020231548473966592,
    228359630832953580969325755111919221821239459840,
    231214126218365500731442327050818212094004953088,
    234068621603777420493558898989717202366770446336,
    236923116989189340255675470928616192639535939584,
    239777612374601260017792042867515182912301432832,
    242632107760013179779908614806414173185066926080,
    245486603145425099542025186745313163457832419328,
    248341098530837019304141758684212153730597912576,
    251195593916248939066258330623111144003363405824,
    254050089301660858828374902562010134276128899072,
    256904584687072778590491474500909124548894392320,
    259759080072484698352608046439808114821659885568,
    262613575457896618114724618378707105094425378816,
    265468070843308537876841190317606095367190872064,
    268322566228720457638957762256505085639956365312,
    271177061614132377401074334195404075912721858560,
    274031556999544297163190906134303066185487351808,
    276886052384956216925307478073202056458252845056,
    279740547770368136687424050012101046731018338304,
    282595043155780056449540621951000037003783831552,
    285449538541191976211657193889899027276549324800,
    288304033926603895973773765828798017549314818048,
    291158529312015815735890337767697007822080311296,
    294013024697427735498006909706595998094845804544,
    296867520082839655260123481645494988367611297792,
    299722015468251575022240053584393978640376791040,
    302576510853663494784356625523292968913142284288,
    305431006239075414546473197462191959185907777536,
    308285501624487334308589769401090949458673270784,
    311139997009899254070706341339989939731438764032,
    313994492395311173832822913278888930004204257280,
    316848987780723093594939485217787920276969750528,
    319703483166135013357056057156686910549735243776,
    322557978551546933119172629095585900822500737024,
    325412473936958852881289201034484891095266230272,
    328266969322370772643405772973383881368031723520,
    331121464707782692405522344912282871640797216768,
    333975960093194612167638916851181861913562710016,
    336830455478606531929755488790080852186328203264,
    339684950864018451691872060728979842459093696512,
    342539446249430371453988632667878832731859189760,
    345393941634842291216105204606777823004624683008,
    348248437020254210978221776545676813277390176256,
    351102932405666130740338348484575803550155669504,
    353957427791078050502454920423474793822921162752,
    356811923176489970264571492362373784095686656000,
    359666418561901890026688064301272774368452149248,
    362520913947313809788804636240171764641217642496,
    365375409332725729550921208179070754913983135744,
    368229904718137649313037780117969745186748628992,
    371084400103549569075154352056868735459514122240,
    373938895488961488837270923995767725732279615488,
    376793390874373408599387495934666716005045108736,
    379647886259785328361504067873565706277810601984,
    382502381645197248123620639812464696550576095232,
    385356877030609167885737211751363686823341588480,
    388211372416021087647853783690262677096107081728,
    391065867801433007409970355629161667368872574976,
    393920363186844927172086927568060657641638068224,
    396774858572256846934203499506959647914403561472,
    399629353957668766696320071445858638187169054720,
    402483849343080686458436643384757628459934547968,
    405338344728492606220553215323656618732700041216,
    408192840113904525982669787262555609005465534464,
    411047335499316445744786359201454599278231027712,
    413901830884728365506902931140353589550996520960,
    416756326270140285269019503079252579823762014208,
    419610821655552205031136075018151570096527507456,
    422465317040964124793252646957050560369293000704,
    425319812426376044555369218895949550642058493952,
    428174307811787964317485790834848540914823987200,
    431028803197199884079602362773747531187589480448,
    433883298582611803841718934712646521460354973696,
    436737793968023723603835506651545511733120466944,
    439592289353435643365952078590444502005885960192,
    442446784738847563128068650529343492278651453440,
    445301280124259482890185222468242482551416946688,
    448155775509671402652301794407141472824182439936,
    451010270895083322414418366346040463096947933184,
    453864766280495242176534938284939453369713426432,
    456719261665907161938651510223838443642478919680,
    459573757051319081700768082162737433915244412928,
    462428252436731001462884654101636424188009906176,
    465282747822142921225001226040535414460775399424,
    468137243207554840987117797979434404733540892672,
    470991738592966760749234369918333395006306385920,
    473846233978378680511350941857232385279071879168,
    476700729363790600273467513796131375551837372416,
    479555224749202520035584085735030365824602865664,
    482409720134614439797700657673929356097368358912,
    485264215520026359559817229612828346370133852160,
    488118710905438279321933801551727336642899345408,
    490973206290850199084050373490626326915664838656,
    493827701676262118846166945429525317188430331904,
    496682197061674038608283517368424307461195825152,
    499536692447085958370400089307323297733961318400,
    502391187832497878132516661246222288006726811648,
    505245683217909797894633233185121278279492304896,
    508100178603321717656749805124020268552257798144,
    510954673988733637418866377062919258825023291392,
    513809169374145557180982949001818249097788784640,
    516663664759557476943099520940717239370554277888,
    519518160144969396705216092879616229643319771136,
    522372655530381316467332664818515219916085264384,
    525227150915793236229449236757414210188850757632,
    528081646301205155991565808696313200461616250880,
    530936141686617075753682380635212190734381744128,
    533790637072028995515798952574111181007147237376,
    536645132457440915277915524513010171279912730624,
    539499627842852835040032096451909161552678223872,
    542354123228264754802148668390808151825443717120,
    545208618613676674564265240329707142098209210368,
    548063113999088594326381812268606132370974703616,
    550917609384500514088498384207505122643740196864,
    553772104769912433850614956146404112916505690112,
    556626600155324353612731528085303103189271183360,
    559481095540736273374848100024202093462036676608,
    562335590926148193136964671963101083734802169856,
    565190086311560112899081243902000074007567663104,
    568044581696972032661197815840899064280333156352,
    570899077082383952423314387779798054553098649600,
    573753572467795872185430959718697044825864142848,
    576608067853207791947547531657596035098629636096,
    579462563238619711709664103596495025371395129344,
    582317058624031631471780675535394015644160622592,
    585171554009443551233897247474293005916926115840,
    588026049394855470996013819413191996189691609088,
    590880544780267390758130391352090986462457102336,
    593735040165679310520246963290989976735222595584,
    596589535551091230282363535229888967007988088832,
    599444030936503150044480107168787957280753582080,
    602298526321915069806596679107686947553519075328,
    605153021707326989568713251046585937826284568576,
    608007517092738909330829822985484928099050061824,
    610862012478150829092946394924383918371815555072,
    613716507863562748855062966863282908644581048320,
    616571003248974668617179538802181898917346541568,
    619425498634386588379296110741080889190112034816,
    622279994019798508141412682679979879462877528064,
    625134489405210427903529254618878869735643021312,
    627988984790622347665645826557777860008408514560,
    630843480176034267427762398496676850281174007808,
    633697975561446187189878970435575840553939501056,
    636552470946858106951995542374474830826704994304,
    639406966332270026714112114313373821099470487552,
    642261461717681946476228686252272811372235980800,
    645115957103093866238345258191171801645001474048,
    647970452488505786000461830130070791917766967296,
    650824947873917705762578402068969782190532460544,
    653679443259329625524694974007868772463297953792,
    656533938644741545286811545946767762736063447040,
    659388434030153465048928117885666753008828940288,
    662242929415565384811044689824565743281594433536,
    665097424800977304573161261763464733554359926784,
    667951920186389224335277833702363723827125420032,
    670806415571801144097394405641262714099890913280,
    673660910957213063859510977580161704372656406528,
    676515406342624983621627549519060694645421899776,
    679369901728036903383744121457959684918187393024,
    682224397113448823145860693396858675190952886272,
    685078892498860742907977265335757665463718379520,
    687933387884272662670093837274656655736483872768,
    690787883269684582432210409213555646009249366016,
    693642378655096502194326981152454636282014859264,
    696496874040508421956443553091353626554780352512,
    699351369425920341718560125030252616827545845760,
    702205864811332261480676696969151607100311339008,
    705060360196744181242793268908050597373076832256,
    707914855582156101004909840846949587645842325504,
    710769350967568020767026412785848577918607818752,
    713623846352979940529142984724747568191373312000,
    716478341738391860291259556663646558464138805248,
    719332837123803780053376128602545548736904298496,
    722187332509215699815492700541444539009669791744,
    725041827894627619577609272480343529282435284992,
    727896323280039539339725844419242519555200778240,
    730750818665451459101842416358141509827966271488,
    733605314050863378863958988297040500100731764736,
    736459809436275298626075560235939490373497257984,
    739314304821687218388192132174838480646262751232,
    742168800207099138150308704113737470919028244480,
    745023295592511057912425276052636461191793737728,
    747877790977922977674541847991535451464559230976,
    750732286363334897436658419930434441737324724224,
    753586781748746817198774991869333432010090217472,
    756441277134158736960891563808232422282855710720,
    759295772519570656723008135747131412555621203968,
    762150267904982576485124707686030402828386697216,
    765004763290394496247241279624929393101152190464,
    767859258675806416009357851563828383373917683712,
    770713754061218335771474423502727373646683176960,
    773568249446630255533590995441626363919448670208,
    776422744832042175295707567380525354192214163456,
    779277240217454095057824139319424344464979656704,
    782131735602866014819940711258323334737745149952,
    784986230988277934582057283197222325010510643200,
    787840726373689854344173855136121315283276136448,
    790695221759101774106290427075020305556041629696,
    793549717144513693868406999013919295828807122944,
    796404212529925613630523570952818286101572616192,
    799258707915337533392640142891717276374338109440,
    802113203300749453154756714830616266647103602688,
    804967698686161372916873286769515256919869095936,
    807822194071573292678989858708414247192634589184,
    810676689456985212441106430647313237465400082432,
    813531184842397132203223002586212227738165575680,
    816385680227809051965339574525111218010931068928,
    819240175613220971727456146464010208283696562176,
    822094670998632891489572718402909198556462055424,
    824949166384044811251689290341808188829227548672,
    827803661769456731013805862280707179101993041920,
    830658157154868650775922434219606169374758535168,
    833512652540280570538039006158505159647524028416,
    836367147925692490300155578097404149920289521664,
    839221643311104410062272150036303140193055014912,
    842076138696516329824388721975202130465820508160,
    844930634081928249586505293914101120738586001408,
    847785129467340169348621865853000111011351494656,
    850639624852752089110738437791899101284116987904,
    853494120238164008872855009730798091556882481152,
    856348615623575928634971581669697081829647974400,
    859203111008987848397088153608596072102413467648,
    862057606394399768159204725547495062375178960896,
    864912101779811687921321297486394052647944454144,
    867766597165223607683437869425293042920709947392,
    870621092550635527445554441364192033193475440640,
    873475587936047447207671013303091023466240933888,
    876330083321459366969787585241990013739006427136,
    879184578706871286731904157180889004011771920384,
    882039074092283206494020729119787994284537413632,
    884893569477695126256137301058686984557302906880,
    887748064863107046018253872997585974830068400128,
    890602560248518965780370444936484965102833893376,
    893457055633930885542487016875383955375599386624,
    896311551019342805304603588814282945648364879872,
    899166046404754725066720160753181935921130373120,
    902020541790166644828836732692080926193895866368,
    904875037175578564590953304630979916466661359616,
    907729532560990484353069876569878906739426852864,
    910584027946402404115186448508777897012192346112,
    913438523331814323877303020447676887284957839360,
    916293018717226243639419592386575877557723332608,
    919147514102638163401536164325474867830488825856,
    922002009488050083163652736264373858103254319104,
    924856504873462002925769308203272848376019812352,
    927711000258873922687885880142171838648785305600,
    930565495644285842450002452081070828921550798848,
    933419991029697762212119024019969819194316292096,
    936274486415109681974235595958868809467081785344,
    939128981800521601736352167897767799739847278592,
    941983477185933521498468739836666790012612771840,
    944837972571345441260585311775565780285378265088,
    947692467956757361022701883714464770558143758336,
    950546963342169280784818455653363760830909251584,
    953401458727581200546935027592262751103674744832,
    956255954112993120309051599531161741376440238080,
    959110449498405040071168171470060731649205731328,
    961964944883816959833284743408959721921971224576,
    964819440269228879595401315347858712194736717824,
    967673935654640799357517887286757702467502211072,
    970528431040052719119634459225656692740267704320,
    973382926425464638881751031164555683013033197568,
    976237421810876558643867603103454673285798690816,
    979091917196288478405984175042353663558564184064,
    981946412581700398168100746981252653831329677312,
    984800907967112317930217318920151644104095170560,
    987655403352524237692333890859050634376860663808,
    990509898737936157454450462797949624649626157056,
    993364394123348077216567034736848614922391650304,
    996218889508759996978683606675747605195157143552,
    999073384894171916740800178614646595467922636800,
    1001927880279583836502916750553545585740688130048,
    1004782375664995756265033322492444576013453623296,
    1007636871050407676027149894431343566286219116544,
    1010491366435819595789266466370242556558984609792,
    1013345861821231515551383038309141546831750103040,
    1016200357206643435313499610248040537104515596288,
    1019054852592055355075616182186939527377281089536,
    1021909347977467274837732754125838517650046582784,
    1024763843362879194599849326064737507922812076032,
    1027618338748291114361965898003636498195577569280,
    1030472834133703034124082469942535488468343062528,
    1033327329519114953886199041881434478741108555776,
    1036181824904526873648315613820333469013874049024,
    1039036320289938793410432185759232459286639542272,
    1041890815675350713172548757698131449559405035520,
    1044745311060762632934665329637030439832170528768,
    1047599806446174552696781901575929430104936022016,
    1050454301831586472458898473514828420377701515264,
    1053308797216998392221015045453727410650467008512,
    1056163292602410311983131617392626400923232501760,
    1059017787987822231745248189331525391195997995008,
    1061872283373234151507364761270424381468763488256,
    1064726778758646071269481333209323371741528981504,
    1067581274144057991031597905148222362014294474752,
    1070435769529469910793714477087121352287059968000,
    1073290264914881830555831049026020342559825461248,
    1076144760300293750317947620964919332832590954496,
    1078999255685705670080064192903818323105356447744,
    1081853751071117589842180764842717313378121940992,
    1084708246456529509604297336781616303650887434240,
    1087562741841941429366413908720515293923652927488,
    1090417237227353349128530480659414284196418420736,
    1093271732612765268890647052598313274469183913984,
    1096126227998177188652763624537212264741949407232,
    1098980723383589108414880196476111255014714900480,
    1101835218769001028176996768415010245287480393728,
    1104689714154412947939113340353909235560245886976,
    1107544209539824867701229912292808225833011380224,
    1110398704925236787463346484231707216105776873472,
    1113253200310648707225463056170606206378542366720,
    1116107695696060626987579628109505196651307859968,
    1118962191081472546749696200048404186924073353216,
    1121816686466884466511812771987303177196838846464,
    1124671181852296386273929343926202167469604339712,
    1127525677237708306036045915865101157742369832960,
    1130380172623120225798162487804000148015135326208,
    1133234668008532145560279059742899138287900819456,
    1136089163393944065322395631681798128560666312704,
    1138943658779355985084512203620697118833431805952,
    1141798154164767904846628775559596109106197299200,
    1144652649550179824608745347498495099378962792448,
    1147507144935591744370861919437394089651728285696,
    1150361640321003664132978491376293079924493778944,
    1153216135706415583895095063315192070197259272192,
    1156070631091827503657211635254091060470024765440,
    1158925126477239423419328207192990050742790258688,
    1161779621862651343181444779131889041015555751936,
    1164634117248063262943561351070788031288321245184,
    1167488612633475182705677923009687021561086738432,
    1170343108018887102467794494948586011833852231680,
    1173197603404299022229911066887485002106617724928,
    1176052098789710941992027638826383992379383218176,
    1178906594175122861754144210765282982652148711424,
    1181761089560534781516260782704181972924914204672,
    1184615584945946701278377354643080963197679697920,
    1187470080331358621040493926581979953470445191168,
    1190324575716770540802610498520878943743210684416,
    1193179071102182460564727070459777934015976177664,
    1196033566487594380326843642398676924288741670912,
    1198888061873006300088960214337575914561507164160,
    1201742557258418219851076786276474904834272657408,
    1204597052643830139613193358215373895107038150656,
    1207451548029242059375309930154272885379803643904,
    1210306043414653979137426502093171875652569137152,
    1213160538800065898899543074032070865925334630400,
    1216015034185477818661659645970969856198100123648,
    1218869529570889738423776217909868846470865616896,
    1221724024956301658185892789848767836743631110144,
    1224578520341713577948009361787666827016396603392,
    1227433015727125497710125933726565817289162096640,
    1230287511112537417472242505665464807561927589888,
    1233142006497949337234359077604363797834693083136,
    1235996501883361256996475649543262788107458576384,
    1238850997268773176758592221482161778380224069632,
    1241705492654185096520708793421060768652989562880,
    1244559988039597016282825365359959758925755056128,
    1247414483425008936044941937298858749198520549376,
    1250268978810420855807058509237757739471286042624,
    1253123474195832775569175081176656729744051535872,
    1255977969581244695331291653115555720016817029120,
    1258832464966656615093408225054454710289582522368,
    1261686960352068534855524796993353700562348015616,
    1264541455737480454617641368932252690835113508864,
    1267395951122892374379757940871151681107879002112,
    1270250446508304294141874512810050671380644495360,
    1273104941893716213903991084748949661653409988608,
    1275959437279128133666107656687848651926175481856,
    1278813932664540053428224228626747642198940975104,
    1281668428049951973190340800565646632471706468352,
    1284522923435363892952457372504545622744471961600,
    1287377418820775812714573944443444613017237454848,
    1290231914206187732476690516382343603290002948096,
    1293086409591599652238807088321242593562768441344,
    1295940904977011572000923660260141583835533934592,
    1298795400362423491763040232199040574108299427840,
    1301649895747835411525156804137939564381064921088,
    1304504391133247331287273376076838554653830414336,
    1307358886518659251049389948015737544926595907584,
    1310213381904071170811506519954636535199361400832,
    1313067877289483090573623091893535525472126894080,
    1315922372674895010335739663832434515744892387328,
    1318776868060306930097856235771333506017657880576,
    1321631363445718849859972807710232496290423373824,
    1324485858831130769622089379649131486563188867072,
    1327340354216542689384205951588030476835954360320,
    1330194849601954609146322523526929467108719853568,
    1333049344987366528908439095465828457381485346816,
    1335903840372778448670555667404727447654250840064,
    1338758335758190368432672239343626437927016333312,
    1341612831143602288194788811282525428199781826560,
    1344467326529014207956905383221424418472547319808,
    1347321821914426127719021955160323408745312813056,
    1350176317299838047481138527099222399018078306304,
    1353030812685249967243255099038121389290843799552,
    1355885308070661887005371670977020379563609292800,
    1358739803456073806767488242915919369836374786048,
    1361594298841485726529604814854818360109140279296,
    1364448794226897646291721386793717350381905772544,
    1367303289612309566053837958732616340654671265792,
    1370157784997721485815954530671515330927436759040,
    1373012280383133405578071102610414321200202252288,
    1375866775768545325340187674549313311472967745536,
    1378721271153957245102304246488212301745733238784,
    1381575766539369164864420818427111292018498732032,
    1384430261924781084626537390366010282291264225280,
    1387284757310193004388653962304909272564029718528,
    1390139252695604924150770534243808262836795211776,
    1392993748081016843912887106182707253109560705024,
    1395848243466428763675003678121606243382326198272,
    1398702738851840683437120250060505233655091691520,
    1401557234237252603199236821999404223927857184768,
    1404411729622664522961353393938303214200622678016,
    1407266225008076442723469965877202204473388171264,
    1410120720393488362485586537816101194746153664512,
    1412975215778900282247703109755000185018919157760,
    1415829711164312202009819681693899175291684651008,
    1418684206549724121771936253632798165564450144256,
    1421538701935136041534052825571697155837215637504,
    1424393197320547961296169397510596146109981130752,
    1427247692705959881058285969449495136382746624000,
    1430102188091371800820402541388394126655512117248,
    1432956683476783720582519113327293116928277610496,
    1435811178862195640344635685266192107201043103744,
    1438665674247607560106752257205091097473808596992,
    1441520169633019479868868829143990087746574090240,
    1444374665018431399630985401082889078019339583488,
    1447229160403843319393101973021788068292105076736,
    1450083655789255239155218544960687058564870569984,
    1452938151174667158917335116899586048837636063232,
    1455792646560079078679451688838485039110401556480,
    1458647141945490998441568260777384029383167049728,
]
